/**********************************************************************
 * File Name: Radsim.java
 *
 * Copyright (C) 2000, US Air Force Research Laboratory,
 * Rome Research Site, Rome NY 13424, USA.
 *
 * Use of this software implies acceptance of the terms described in
 * accompanying license.txt file.
 *
 *
 * Authors: James Lawton, Jennifer Dzek, John Crowter,
 *          Bob Paragi, and Joe Beasock
 *
 * Creation Date: July 1999
 *
 * Description: The main class for the RadSim simulation.
 *
 * Modification History:
 *
 *  - June 00, Bryan Horling, UMass (BCH): Changes to make drawing more 
 *  efficient by only redrawing when needed.
 *
 **********************************************************************/

import java.awt.*;
import java.lang.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.applet.Applet;
import java.util.*;
import java.io.*;

public class Radsim implements Runnable {

    public SimEnvironment SimEnv;
    SimClock Clock;
    public Thread simThread;
    public static RadsimFrame DrawingFrame;
    private String configFile = "config.txt";
    private String logFile = "radsim-log.txt";
    private int debug = 0;
    private int loglevel = 0;
    private boolean showGraphics = true;
    private int minimum_realtime = 10;
    private int maximum_simtime = Integer.MAX_VALUE;
    private int mobj_start_time = 0; //When to start the mobile objects moving

    /**
     * Initialize the application.
     */
    public Radsim(String args[]) throws Exception{

	if (! parseArgs(args) )
	    throw(new Exception("Usage: Radsim [-d debug_level] [-c config_file] [-ng] " +
				"[-m min_realtim] [-s max_simtime] " +
				"[-l loglevel [-lf logfile]] [-t target_start_time]"));
	
	// Set up the simulation environment object:
	SimEnv = new SimEnvironment(100, 100);
	SimEnv.Debug = debug;
	
	// Set u the simulation clock:
	Clock = new SimClock();
	Clock.env = SimEnv;
	SimEnv.clock = Clock;

	// Set the real-to-simulation time ratio
	if (minimum_realtime > 0)
	    SimEnv.time_ratio = (float) minimum_realtime / (float) Clock.ticksize;

	// Set up logging
	if (loglevel > 0) {
	    Log.init(logFile, loglevel);
	    Log.timestamp(0);
	}

	// Read in config file and create the simulation objects:
	Sensor newSensor = null;
	MobileObject newMObj = null;
	boolean OK = true, done = false;
	BufferedReader instream = null;
	String line = null;

	try { instream = new BufferedReader(new FileReader(configFile)); }
	catch (Exception e) {
	    System.out.println(e);
	    done = true;
	}

	while (! done) {
	    OK = true;
	    try {line = instream.readLine();}
	    catch (Exception e) {
		System.out.println(e);
		done = true;
	    }

	    if (done || line == null) {
		done = true;
	    }
	    else if (line.equals("") || line.startsWith("#")) { // skip it
	    }
	    else if (line.startsWith("ROOMSIZE")) {
		String sizestring = line.substring(9); //trim off the tag
		try {
		    int ind = sizestring.indexOf(" ");
		    int new_width = (new Integer(sizestring.substring(0, ind))).intValue();
		    int new_height = (new Integer(sizestring.substring(ind + 1))).intValue();
		    SimEnv.width = new_width;
		    SimEnv.height = new_height;
		    Log.roomsize(new_width, new_height);
		}
		catch (Exception e) {
		    System.out.println("  Error setting roomsize: " + e);
		}
	    }
	    else if (line.startsWith("SENSOR")) {
		try {newSensor = new Sensor(line);}
		catch (Exception e) {
		    System.out.println("  Sensor Constructor error: " + e);
		    OK = false;
		}
		if (OK) {
		    SimEnv.add_sensor(newSensor);
		    SimEnv.commChannels[newSensor.recChannel].register(newSensor);
		    newSensor.initSensor();
		    // newSensor.schedule(newSensor.GET_COMMAND_EVENT,0);
		    if (SimEnv.Debug > 1)
			System.out.println("New sensor created -- Name: " + newSensor.name);
		}
	    }
	    else if (line.startsWith("MOBOBJ") || line.startsWith("MODOBJ")) {
		try {
		    if (line.startsWith("MOBOBJ")) {
			newMObj = new MobileObject(line, SimEnv.width, SimEnv.height);
		    }
		    else {
			newMObj = new MobileDirectedObject(line, SimEnv.width, SimEnv.height);
		    }
		}
		catch (Exception e) {
		    System.out.println("  MobObj Constructor error: " + e);
		    OK = false;
		}
		if (OK) {
		    SimEnv.add_mobile_object(newMObj);
		    newMObj.initMobileObject(mobj_start_time * 1000 / Clock.ticksize);
		    //newMObj.schedule(newMObj.TICK_EVENT, mobj_start_time * 1000 / Clock.ticksize);
		    //newMObj.last_update = mobj_start_time * 1000 / Clock.ticksize;
		    if (SimEnv.Debug > 1)
			System.out.println("New MobileObject created -- Name: " + newMObj.name);
		}
	    }
	    else
		System.out.println("Unknown line: " + line);
	}

	// Fire up the graphics:
	if (showGraphics) {
	    try {
		DrawingFrame = new RadsimFrame();
		DrawingFrame.initComponents(SimEnv);
		SimEnv.Display = DrawingFrame.drawingCanvas;
		SimEnv.Status = DrawingFrame.statusCanvas;
		DrawingFrame.setVisible(true);
	    }
	    catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    public void radStart() {
	simThread = new Thread(this);
	simThread.start();
    }

    public void stop() {
	//	simThread.stop();
    }

    /**
     * Cleans up whatever resources are being held.  If the applet is active
     * it is stopped.
     */
    public void destroy() {
    }

    public void run() {
	long start_time, elapsed_time;
	long sim_start_time, sim_end_time;

	// Adjust the max simtime to clock ticks.
	if (maximum_simtime < Integer.MAX_VALUE)
	    maximum_simtime = (maximum_simtime * 1000) / Clock.ticksize;

	sim_start_time = System.currentTimeMillis();

	while((! SimEnv.quitCalled) && (Clock.simtime < maximum_simtime)) {
	    if (SimEnv.running) {
		start_time = System.currentTimeMillis();
		Clock.run_for(1);
		
		if (showGraphics)
		    SimEnv.update_display();

		elapsed_time = System.currentTimeMillis() - start_time;
		
		if (elapsed_time < minimum_realtime) {
		    try{ Thread.sleep(minimum_realtime - elapsed_time); } 
		    catch ( Exception e ) {};
		}
	    } 
	    // Make sure things don't lock up when the simulation is stopped
	    // Bryan Horling (UMass)
	    else {
		try{ Thread.sleep(500); } 
		catch ( Exception e ) {};
		if (showGraphics)
		    SimEnv.update_display();
	    }
	}

	// Gather and print out timing information:
	sim_end_time = System.currentTimeMillis();
	System.out.println("Elapsed real time (milliseconds) = " + (sim_end_time - sim_start_time));
	System.out.println("Elapsed simtime (milliseconds) = " + (Clock.simtime * Clock.ticksize));
	System.out.println("Real time to simtime ratio = " + 
			   ((double) (sim_end_time - sim_start_time) /
			    (Clock.simtime * Clock.ticksize) ));


	// Cleanly shut down all the objects:
	for (ListIterator objs = SimEnv.sim_objects.listIterator(); objs.hasNext(); ) {
	    SimObject obj = (SimObject) objs.next();
	    obj.shutdown();
	}
	
	// Shut down the logging facility
	Log.timestamp(Clock.simtime);
	Log.shutdown();

	System.exit(0);
    }

    private boolean parseArgs(String args[]) {
	int i;
	boolean argsOK = true;
	if (args.length > 0) {
	    for(i = 0; i < args.length && argsOK; ) {
		if (args[i].equals("-c")) {
		    if (i+1 < args.length) {
			configFile = args[i+1];
			System.out.println("Using config file " + configFile);
			i = i + 2;
		    }
		    else
			argsOK = false;
		}
		else if (args[i].equals("-d")) {
		    if (i+1 < args.length) {
			try {
			    debug = (new Integer(args[i+1])).intValue();
			    System.out.println("Setting debug level to " + debug);
			    i = i + 2;
			}
			catch (Exception e) {
			    argsOK = false;
			}
		    }
		    else
			argsOK = false;
		}
		else if (args[i].equals("-m")) {
		    if (i+1 < args.length) {
			try {
			    minimum_realtime = (new Integer(args[i+1])).intValue();
			    System.out.println("Setting minimum real time per clock tick to " 
					       + minimum_realtime + " milliseconds");
			    i = i + 2;
			}
			catch (Exception e) {
			    argsOK = false;
			}
		    }
		    else
			argsOK = false;
		}
		else if (args[i].equals("-s")) {
		    if (i+1 < args.length) {
			try {
			    maximum_simtime = (new Integer(args[i+1])).intValue();
			    System.out.println("Setting maximum simtime to " + maximum_simtime +
					       " seconds");
			    i = i + 2;
			}
			catch (Exception e) {
			    argsOK = false;
			}
		    }
		    else
			argsOK = false;
		}

		else if (args[i].equals("-t")) {
		    if (i+1 < args.length) {
			try {
			    mobj_start_time = (new Integer(args[i+1])).intValue();
			    System.out.println("Setting target start time to " + mobj_start_time);
			    i = i + 2;
			}
			catch (Exception e) {
			    argsOK = false;
			}
		    }
		    else
			argsOK = false;
		}


		else if (args[i].equals("-l")) {
		    if (i+1 < args.length) {
			try {
			    int lvl = (new Integer(args[i+1])).intValue();
			    if (lvl >= 0) {
				System.out.println("Setting logging level to " + lvl);
				loglevel = lvl;
			    }
			    else {
				System.out.println("Invalid logging level: " + lvl);
				argsOK = false;
			    }

			    i = i + 2;
			}
			catch (Exception e) {
			    argsOK = false;
			}
		    }
		    else
			argsOK = false;
		}

		else if (args[i].equals("-lf")) {
		    if (i+1 < args.length) {
			logFile = args[i+1];
			System.out.println("Using log file " + logFile);
			i = i + 2;
		    }
		    else
			argsOK = false;
		}

		else if (args[i].equals("-ng")) {
		    showGraphics = false;
		    System.out.println("Running without graphics");
		    i = i + 1;
		}

		else
		    argsOK = false;
	    }
	}
	return(argsOK);
    }



    public static void main( String args[] ) {
	try {
	    Radsim RadSim = new Radsim(args);
	    RadSim.SimEnv.running = true;     //  added in application form of this module
	    RadSim.radStart();
	}
	catch(Exception e) {
	    System.out.println(e);
	}
    }
}

